import re
from typing import Dict, Literal, Optional, Union
from uuid import UUID

from grai_schemas.generics import GraiBaseModel
from grai_schemas.v1.organization import OrganisationSpec, OrganisationV1
from pydantic import Field, validator


class _DefaultRefSentinel(str):
    pass


_REF_REGEX = re.compile("^[^/]*/[^/]*$")


class WorkspaceSpec(GraiBaseModel):
    """Class definition of WorkspaceSpec

    Attributes:
        id: Optional UUID of the workspace
        name: The name of the workspace
        organisation: The organisation the workspace belongs to
        ref: The reference of the workspace in the form of `organisation/name`
        search_enabled: Whether the workspace is searchable or not

    """

    id: Optional[UUID]
    name: str
    organisation: Union[UUID, OrganisationSpec] = Field(..., alias="organization")
    ref: str = _DefaultRefSentinel()  # This keeps mypy happy
    search_enabled: Optional[bool]

    @property
    def organization(self):
        """Alias for organisation"""
        return self.organisation

    @validator("organisation", always=True, pre=True)
    def validate_organisation(cls, v: Union[UUID, str, Dict, OrganisationSpec]) -> Union[UUID, OrganisationSpec]:
        if isinstance(v, (OrganisationSpec, UUID)):
            return v
        elif isinstance(v, OrganisationV1):
            return v.spec
        elif isinstance(v, dict):
            return OrganisationSpec(**v)
        elif isinstance(v, str):
            try:
                uuid_val = UUID(v)
                return uuid_val
            except ValueError:
                return OrganisationSpec(name=v)
        else:
            message = (
                f"Invalid value for organisation: Received {v} which is of type {type(v)}. You might have meant "
                f"to provide a UUID for the organisation, or a string containing the name of the organisation?"
            )
            raise ValueError(message)

    @validator("ref", always=True, pre=True)
    def validate_ref(cls, v: Optional[str], values, field) -> str:
        if v is None or isinstance(v, _DefaultRefSentinel):
            if isinstance(values.get("organisation", None), OrganisationSpec):
                return f"{values['organisation'].name}/{values['name']}"
            else:
                message = (
                    f"`ref` is a required field when `organisation` is a UUID. Either provide a value for `ref`, or "
                    f" provide an OrganizationSpec to organisation and ref will be automatically generated."
                )
                raise ValueError(message)
        elif re.match(_REF_REGEX, v):
            # TODO: This doesn't cover the case of a user providing a UUID for the organisation
            if isinstance(values.get("organisation", None), OrganisationSpec):
                validated_ref = f"{values['organisation'].name}/{values['name']}"
                message = (
                    f"Attempted to create a workspace with a `ref` of {v}, but this didn't match the specified "
                    f"organisation name ({values['organisation'].name}) and workspace name ({values['name']}). "
                )
                if validated_ref != v:
                    raise ValueError(message)
            return v
        else:
            message = (
                f"Invalid value for `ref`: Received {v}. Expected a string containing a"
                f"single forwards slash (/) separating the organisation name and the workspace name."
                f"e.g. `my-organisation/my-workspace`"
            )
            raise ValueError(message)

    def __hash__(self) -> int:
        """Workspace hash definition"""
        return hash(self.ref)


class WorkspaceV1(GraiBaseModel):
    """Class definition of WorkspaceV1

    Attributes:
        type: todo
        version: todo
        spec: todo

    """

    type: Literal["Workspace"] = "Workspace"
    version: Literal["v1"] = "v1"
    spec: WorkspaceSpec

    @classmethod
    def from_spec(cls, spec_dict: Dict) -> "WorkspaceV1":
        """

        Args:
            spec_dict (Dict):

        Returns:

        Raises:

        """
        return cls(version="v1", type="Workspace", spec=spec_dict)

    def __hash__(self) -> int:
        return hash(self.spec)


__all__ = ["WorkspaceSpec", "WorkspaceV1"]



from typing import Any, Dict, Literal, Optional, TypeVar, Union
from uuid import UUID

from grai_schemas.serializers import dump_json, load_json
from grai_schemas.utilities import merge
from pydantic import BaseModel, dataclasses, root_validator, validator

T = TypeVar("T")


class HashableBaseModel(BaseModel):
    """A BaseModel that is hashable"""

    def __hash__(self):
        return id(self)


class GraiBaseModel(HashableBaseModel):
    """The base class for all Grai models

    This class provides a number of features which are useful for Grai models:
    * hashable - this allows Grai models to be used as keys in dictionaries
    * update - this allows Grai models to be updated with new values
    * json_loads - this allows Grai models to be loaded from JSON
    * json_dumps - this allows Grai models to be dumped to JSON

    In addition there is are pydantic specific configuration changes which enforce consistent behavior across Grai Models:
    * validate_all - this ensures that all fields are validated
    * validate_assignment - this ensures that all fields are validated when assigned
    * allow_population_by_field_name - this allows Grai models to be updated with new values by field name
    * orm_mode - this allows Grai models to be used with ORMs

    """

    # class Config:
    #     frozen = True

    def update(self, new_values: Dict) -> BaseModel:
        """Automatically update a Grai model with new values

        Update uses the `merge` function to update the current model with new values.
        Merge understands the nested structure of Grai models and will update nested models correctly.

        Args:
            new_values (Dict):

        Returns:
            An updated instance of the current model

        Raises:

        """
        values = self.dict()
        return type(self)(**merge(values, new_values))

    class Config:
        """ """

        json_loads = load_json
        json_dumps = dump_json
        # json_encoders = {UUID: lambda x: str(x)}
        validate_all = True
        validate_assignment = True
        allow_population_by_field_name = True
        orm_mode = True


class PlaceHolderSchema(GraiBaseModel):
    """Class definition of PlaceHolderSchema

    This is a placeholder schema which is used when a schema version is not yet available.
    It should not be used for any other purpose.

    Attributes:
        is_active: todo

    """

    is_active: Optional[bool] = True

    @root_validator(pre=True)
    def root_validator_of_placeholder(cls, values):
        """

        Args:
            values:

        Returns:

        Raises:

        """
        message = (
            "Something is wrong... I can feel it 😡. You've reached a placeholder schema - "
            "most likely the `version` of your config file doesn't exist yet."
        )
        raise AssertionError(message)


# ----


class DefaultValue(GraiBaseModel):
    """Class definition of DefaultValue

    Attributes:
        has_default_value: Identifies whether a default value is available
        data_type: The data type of the default value
        default_value: The default value

    """

    has_default_value: Optional[bool] = None
    data_type: Optional[str] = None
    default_value: Optional[Any] = None

    @root_validator()
    def validate_default_value_root(cls, values):
        """

        Args:
            values:

        Returns:

        Raises:

        """
        if isinstance(values, dict):
            has_default_value = values.get("has_default_value", None)
            data_type = values.get("data_type", None)
            default_value = values.get("default_value", None)
        else:
            raise NotImplementedError(f"No available implementation to produce a DefaultValue from a {type(values)}")

        if has_default_value is None or has_default_value is False:
            assert data_type is None, "Cannot set a data_type when `has_default_value` is not True"
            assert default_value is None, "Cannot set a default_value when `has_default_value` is not True"
        else:
            assert data_type is not None, "If `has_default_value` is True, a `data_type` is required"
            assert default_value is not None, "If `has_default_value` is True, a `default_value` is required"

        return values


# ----


class PackageConfig(BaseModel):
    """Class definition of PackageConfig

    Attributes:
        integration_name: todo
        metadata_id: todo

    """

    integration_name: str
    metadata_id: str

    @validator("metadata_id")
    def metadata_id_validation(cls, value):
        """

        Args:
            value:

        Returns:

        Raises:

        """
        assert (
            "-" not in value
        ), f"Error found in metadata_id: {value}. `-` is a reserved character which should not be used."
        return value

    class Config:
        """ """

        validate_assignment = True
        validate_all = True


class Metadata(GraiBaseModel):
    """A base class for all metadata models"""

    pass

    class Config:
        """ """

        extra = "allow"
        allow_mutation = True


class MalformedMetadata(GraiBaseModel):
    """Class definition of MalformedMetadata

    Attributes:
        malformed_values: A cache of values used to instantiate the class.

    """

    malformed_values: Dict = {}

    @root_validator(pre=True)
    def validate_malformed(cls, v):
        """ """
        return {"malformed_values": {**v}}

    def dict(self, *args, **kwargs):
        """ """
        return self.malformed_values



import abc
from typing import Annotated, Dict, Literal, Optional, Set, Union, get_args
from uuid import UUID

from grai_schemas.generics import GraiBaseModel, Metadata
from pydantic import BaseModel, Field, root_validator, validator


class BaseID(GraiBaseModel):
    """Class definition of BaseID

    Attributes:
        id: Optional UUID of the object
        name: Optional name of the object
        namespace: Optional namespace of the object

    """

    id: Optional[UUID]
    name: Optional[str]
    namespace: Optional[str]

    def __hash__(self):
        if self.name is None or self.namespace is None:
            raise NotImplementedError(
                f"Computing a hash for {self} requires both name and namespace for compatibility reasons."
            )
        return hash((self.name, self.namespace))


class NamedID(BaseID):
    """Class definition of NamedID

    Attributes:
        id: Optional UUID of the object
        name: Name of the object
        namespace: Namespace of the object

    """

    id: Optional[UUID]
    name: str
    namespace: str


class UuidID(BaseID):
    """Class definition of UuidID

    Attributes:
        id: UUID of the object
        name: Optional name of the object
        namespace: Optional namespace of the object

    """

    id: UUID
    name: Optional[str]
    namespace: Optional[str]


ID = Union[UuidID, NamedID]


class ProgrammingLanguage(BaseModel):
    pass


class Python(ProgrammingLanguage):
    """Class representation of the Python programming language"""

    language_name: Literal["Python"] = "Python"
    file_extension: Literal[".py"] = ".py"


class R(ProgrammingLanguage):
    """Class representation of the R programming language"""

    language_name: Literal["R"] = "R"
    file_extension: Literal[".r"] = ".r"


class SQL(ProgrammingLanguage):
    """Class representation of the SQL programming language"""

    language_name: Literal["SQL"] = "SQL"
    file_extension: Literal[".sql"] = ".sql"


class C(ProgrammingLanguage):
    """Class representation of the C programming language"""

    language_name: Literal["C"] = "C"
    file_extension: Literal[".c"] = ".c"


class CSharp(ProgrammingLanguage):
    """Class representation of the C# programming language"""

    language_name: Literal["C#"] = "C#"
    file_extension: Literal[".cs"] = ".cs"


class CPP(ProgrammingLanguage):
    """Class representation of the SQL programming language"""

    language_name: Literal["C++"] = "C++"
    file_extension: Literal[".c"] = ".c"


class Java(ProgrammingLanguage):
    """Class representation of the Java programming language"""

    language_name: Literal["Java"] = "Java"
    file_extension: Literal[".java"] = ".java"


class Scala(ProgrammingLanguage):
    """Class representation of the Scala programming language"""

    language_name: Literal["Scala"] = "Scala"
    file_extension: Literal[".scala", ".sc"] = ".scala"


class Go(ProgrammingLanguage):
    """Class representation of the Go programming language"""

    language_name: Literal["Go"] = "Go"
    file_extension: Literal[".go"] = ".go"


class JavaScript(ProgrammingLanguage):
    """Class representation of the JavaScript programming language"""

    language_name: Literal["JavaScript"] = "JavaScript"
    file_extension: Literal[".js"] = ".js"


class TypeScript(ProgrammingLanguage):
    """Class representation of the TypeScript programming language"""

    language_name: Literal["TypeScript"] = "TypeScript"
    file_extension: Literal[".ts"] = ".ts"


class Matlab(ProgrammingLanguage):
    """Class representation of the Matlab programming language"""

    language_name: Literal["Matlab"] = "Matlab"
    file_extension: Literal[".m"] = ".m"


class Swift(ProgrammingLanguage):
    """Class representation of the Swift programming language"""

    language_name: Literal["Swift"] = "Swift"
    file_extension: Literal[".swift"] = ".swift"


class Julia(ProgrammingLanguage):
    """Class representation of the Julia programming language"""

    language_name: Literal["Julia"] = "Julia"
    file_extension: Literal[".jl"] = ".jl"


class SAS(ProgrammingLanguage):
    """Class representation of the SAS programming language"""

    language_name: Literal["SAS"] = "SAS"
    file_extension: Literal[".sas"] = ".sas"


class Rust(ProgrammingLanguage):
    """Class representation of the Rust programming language"""

    language_name: Literal["Rust"] = "Rust"
    file_extension: Literal[".rs"] = ".rs"


class Perl(ProgrammingLanguage):
    """Class representation of the Perl programming language"""

    language_name: Literal["Perl"] = "Perl"
    file_extension: Literal[".pl"] = ".pl"


class Haskell(ProgrammingLanguage):
    """Class representation of the Haskell programming language"""

    language_name: Literal["Haskell"] = "Haskell"
    file_extension: Literal[".hs"] = ".hs"


class PHP(ProgrammingLanguage):
    """Class representation of the PHP programming language"""

    language_name: Literal["PHP"] = "PHP"
    file_extension: Literal[".php"] = ".php"


class Kotlin(ProgrammingLanguage):
    """Class representation of the Kotlin programming language"""

    language_name: Literal["Kotlin"] = "Kotlin"
    file_extension: Literal[".kt"] = ".kt"


class UnknownLanguage(ProgrammingLanguage):
    """Class representation of a catch-all programming language"""

    language_name: Literal["Unknown"] = "Unknown"
    original_language_name: Optional[str] = None
    file_extension: Optional[str] = None

    @root_validator(pre=True)
    def validate_malformed(cls, v: Dict):
        if v.get("language_name", "Unknown") != "Unknown":
            v["original_language_name"] = v["language_name"]
            v["language_name"] = "Unknown"
        return v


# Unknown must be last in the list as it's a catch-all
Languages = Union[
    Python,
    R,
    SQL,
    C,
    CPP,
    CSharp,
    Java,
    Scala,
    Go,
    JavaScript,
    TypeScript,
    Matlab,
    Swift,
    Julia,
    SAS,
    Rust,
    Haskell,
    Perl,
    PHP,
    Kotlin,
    UnknownLanguage,
]
LANGUAGE_LABELS: Set[str] = set()
for model in get_args(Languages):
    field_type = model.__annotations__["language_name"]
    LANGUAGE_LABELS.update(get_args(field_type))


ProgrammingLanguageTypes = Annotated[Languages, Field(discriminator="language_name")]


class Code(BaseModel):
    """A generic descriptor for Code"""

    code: Optional[str]
    has_code: bool = False
    language: ProgrammingLanguageTypes = UnknownLanguage()

    @root_validator(pre=True)
    def validate_root(cls, values):
        language = values.get("language", None)
        if isinstance(language, str):
            if language not in LANGUAGE_LABELS:
                values["language"] = {"language_name": "Unknown", "original_language_name": language}
            else:
                values["language"] = {"language_name": language}
        return values

    @validator("has_code", pre=True, always=True)
    def set_has_code(cls, value, values):
        return bool(values.get("code"))



from typing import Literal, Optional

from grai_schemas import generics


class GenericAttributes(generics.HashableBaseModel):
    """Class definition of GenericAttributes"""

    class Config:
        """ """

        extra = "allow"
        allow_mutation = True



from typing import Literal, Optional
from uuid import UUID

from grai_schemas.generics import GraiBaseModel
from pydantic import validator


class OrganisationSpec(GraiBaseModel):
    """Class definition of OrganisationSpec

    Attributes:
        name: The name of the organisation.
        id: Optional UUID of the organisation

    """

    name: str
    id: Optional[UUID]

    def __hash__(self) -> int:
        return hash(self.name)


class OrganisationV1(GraiBaseModel):
    """Class definition of OrganisationV1

    Attributes:
        type: The type of the object e.g. Node, Edge, etc.
        version: The version of the object e.g. v1
        spec: The specification of the object.

    """

    type: Literal["Organisation"]
    version: Literal["v1"] = "v1"
    spec: OrganisationSpec

    @validator("type")
    def validate_org_type(cls, v):
        if v == "Organization":
            return "Organisation"
        return v

    def __hash__(self) -> int:
        return hash(self.spec)

    @classmethod
    def from_spec(cls, spec_dict: dict) -> "OrganisationV1":
        """

        Args:
            spec_dict (Dict):

        Returns:

        Raises:

        """
        return cls(version="v1", type="Organisation", spec=spec_dict)


__all__ = ["OrganisationSpec", "OrganisationV1"]



import copy
from typing import Any, Dict, Optional, Union

from grai_schemas.generics import MalformedMetadata, Metadata
from grai_schemas.utilities import merge
from grai_schemas.v1.edge import EdgeSpec, EdgeV1, SourcedEdgeSpec, SourcedEdgeV1
from grai_schemas.v1.generics import Code
from grai_schemas.v1.metadata.edges import BaseEdgeMetadataV1
from grai_schemas.v1.metadata.nodes import BaseNodeMetadataV1
from grai_schemas.v1.node import NodeSpec, NodeV1, SourcedNodeSpec, SourcedNodeV1

MetadataType = Union[BaseNodeMetadataV1, BaseEdgeMetadataV1]


@merge.register
def merge_malformed_left(metadata: MalformedMetadata, other_metadata: Any) -> Metadata:
    """Merge anything into a malformed metadata object

    This handles cases where one might attempt to fix a malformed metadata object by merging a valid metadata object.

    Args:
        metadata: The malformed metadata to merge into
        other_metadata: The metadata to merge from

    Returns:
        A now valid piece of metadata
    """
    return other_metadata


@merge.register
def merge_malformed_right(metadata: Any, other_metadata: MalformedMetadata):
    """Merge a malformed metadata object into anything

    Args:
        metadata: The metadata to merge into
        other_metadata: The malformed metadata to merge from

    Raises:
        ValueError: This is always an invalid operation
    """
    raise ValueError(
        "Cannot merge malformed metadata into valid metadata. If you received this error check if any of the metadata"
        "in your objects is of type `MalformedMetadata`. This could indicate a bad value has been stored in your"
        "instance of Grai."
    )


def merge_tags(a: Optional[list], b: Optional[list]) -> list:
    """Merge two lists of tags insuring no duplicates"""
    a_tag = set(a) if a is not None else set()
    b_tag = set(b) if b is not None else set()
    return list(a_tag.union(b_tag))


@merge.register
def merge_grai_node_v1_metadata(metadata: BaseNodeMetadataV1, other_metadata: BaseNodeMetadataV1) -> BaseNodeMetadataV1:
    """Merge two grai node metadata objects

    Args:
        metadata: The node metadata to merge into
        other_metadata: The node metadata to merge from

    Returns:
        The merged node metadata
    """
    new_metadata = merge(dict(metadata), dict(other_metadata))
    new_metadata["tags"] = merge_tags(metadata.tags, other_metadata.tags)
    return BaseNodeMetadataV1(**new_metadata)


@merge.register
def merge_grai_edge_v1_metadata(metadata: BaseEdgeMetadataV1, other_metadata: BaseEdgeMetadataV1) -> BaseEdgeMetadataV1:
    """Merge two grai edge metadata objects

    Args:
        metadata: The edge metadata to merge into
        other_metadata: The edge metadata to merge from

    Returns:
        The merged edge metadata
    """
    new_metadata = merge(dict(metadata), dict(other_metadata))
    new_metadata["tags"] = merge_tags(metadata.tags, other_metadata.tags)
    return BaseEdgeMetadataV1(**new_metadata)


@merge.register
def merge_node_sourced_node(node: NodeV1, source_node: SourcedNodeV1) -> NodeV1:
    """Merge a sourced node into a node

    Args:
        node: The node to merge into
        source_node: The sourced node to merge from

    Returns:
        The merged NodeV1

    """
    new_node = node.copy()
    new_node.spec = merge(node.spec, source_node.spec)
    return new_node


@merge.register
def merge_edge_sourced_edge(edge: EdgeV1, source_edge: SourcedEdgeV1) -> EdgeV1:
    """Merge a sourced edge into an edge

    Args:
        edge: The edge to merge into
        source_edge: The sourced edge to merge from

    Returns:
        The merged EdgeV1

    """
    new_edge = edge.copy()
    new_edge.spec = merge(edge.spec, source_edge.spec)
    return new_edge


@merge.register
def merge_grai_node_into_node_metadata(spec: NodeSpec, source_spec: SourcedNodeSpec) -> NodeSpec:
    """Merge a SourcedNodeSpec into a NodeSpec

    Args:
        spec: The node spec to merge into
        source_spec: The sourced node spec to merge from

    Returns:
        The merged node spec

    """
    new_spec = copy.deepcopy(spec)
    new_spec.metadata.grai = merge(new_spec.metadata.grai, source_spec.metadata.grai)
    new_spec.metadata.sources[source_spec.data_source.name] = source_spec.metadata
    return new_spec


@merge.register
def merge_grai_edge_into_edge_metadata(spec: EdgeSpec, source_spec: SourcedEdgeSpec) -> EdgeSpec:
    """Merge grai metadata from a sourced edge into an edge spec

    Args:
        spec: The edge spec to merge into
        source_spec: The sourced edge spec to merge from

    Returns:
        The merged edge spec
    """
    new_spec = copy.deepcopy(spec)
    new_spec.metadata.grai = merge(new_spec.metadata.grai, source_spec.metadata.grai)
    new_spec.metadata.sources[source_spec.data_source.name] = source_spec.metadata
    return new_spec


@merge.register
def merge_code_into_code(code_a: Code, code_b: Code) -> Code:
    return code_b



from typing import Dict, Literal, Union

from grai_schemas.base import GraiType
from grai_schemas.generics import GraiBaseModel


class Schema(GraiBaseModel):
    """Class definition of Schema

    Attributes:
        entity: A Grai object

    """

    entity: GraiType

    @classmethod
    def to_model(cls, item: Dict, version: Literal["v1"], typing_type: Literal["Node", "Edge"]) -> GraiType:
        """Convert an item spec to a Grai object

        Args:
            item: An item spec to be converted to a Grai object
            version: which version of the schema to use
            typing_type: The type of the object e.g. Node, Edge, etc.

        Returns:
            The Grai object

        Raises:

        """
        result = {
            "type": typing_type,
            "version": version,
            "spec": item,
        }
        return cls(entity=result).entity



from typing import List, Literal, Optional, Union
from uuid import UUID

from grai_schemas.generics import GraiBaseModel
from grai_schemas.v1.workspace import WorkspaceSpec
from pydantic import validator


class SourceSpec(GraiBaseModel):
    """Class definition of SourceSpec

    Attributes:
        id: An optional UUID of the source.
        name: The name of the source.
        workspace: The workspace the source belongs to.

    """

    id: Optional[UUID]
    name: str
    workspace: Union[UUID, WorkspaceSpec, None]

    @property
    def workspace_id(self) -> Optional[UUID]:
        return self.workspace.id if isinstance(self.workspace, WorkspaceSpec) else self.workspace

    def __hash__(self) -> int:
        return hash(self.name)


class DataSourceMixin(GraiBaseModel):
    """Class definition of DataSourceMixin

    Attributes:
        data_source: The data source which created this object.

    """

    data_source: SourceSpec


class DataSourcesMixin(GraiBaseModel):
    """Class definition of DataSourcesMixin

    Attributes:
        data_sources: The data sources which created this object.

    """

    data_sources: List[Union[UUID, SourceSpec]]


class SourceV1(GraiBaseModel):
    """Class definition of SourceV1

    Attributes:
        type: A string indicating the type of the object. In this case it is "Source".
        version: The version of the object e.g. "v1".
        spec: The specification of the object.

    """

    type: Literal["Source"] = "Source"
    version: Literal["v1"] = "v1"
    spec: SourceSpec

    def __hash__(self):
        return hash(self.spec)

    @classmethod
    def from_spec(cls, spec: Union[dict, SourceSpec]) -> "SourceV1":
        """

        Args:
            spec: The specification of the object.

        Returns:

        Raises:

        """
        return cls(version="v1", type="Source", spec=spec)


__all__ = ["SourceSpec", "SourceV1", "DataSourceMixin", "DataSourcesMixin"]



from abc import ABC, abstractmethod
from typing import List, Literal, Set

from grai_schemas.base import Event, SourcedEdge, SourcedNode
from pydantic import BaseModel


class QuarantineReason(BaseModel, ABC):
    """Base class for quarantine reasons"""

    @property
    @abstractmethod
    def reason(self) -> str:
        """Returns a string describing the reason for quarantine"""
        pass


class MissingEdgeNodeReason(QuarantineReason):
    """Class definition of MissingEdgeNodeReason

    Attributes:
        side: Either Source or Destination
        node_name: The name of the missing node
        node_namespace: The namespace of the missing node

    """

    side: Literal["source", "destination"]
    node_name: str
    node_namespace: str

    @property
    def reason(self) -> str:
        """Returns a string describing the reason for quarantine"""
        return f"{self.side.capitalize()} node `({self.node_name}, {self.node_namespace})` was not found."


class QuarantinedEdge(BaseModel):
    """Class definition of QuarantinedEdge

    Attributes:
        edge: The edge that was quarantined
        reasons: A list of reasons for quarantine

    """

    edge: SourcedEdge
    reasons: List[QuarantineReason]


class QuarantinedNode(BaseModel):
    """Class definition of QuarantinedNode

    Attributes:
        node: The node that was quarantined
        reasons: A list of reasons for quarantine

    """

    node: SourcedNode
    reasons: List[QuarantineReason]


class QuarantinedEvent(BaseModel):
    """Class definition of QuarantinedEvent

    Attributes:
        event: The event that was quarantined
        reasons: A list of reasons for quarantine

    """

    event: Event
    reasons: List[QuarantineReason]


class Quarantine(BaseModel):
    """Class definition of Quarantine

    Attributes:
        nodes: A list of quarantined nodes
        edges: A list of quarantined edges
        events: A list of quarantined events

    """

    nodes: List[QuarantinedNode] = []
    edges: List[QuarantinedEdge] = []
    events: List[QuarantinedEvent] = []

    @property
    def has_quarantined(self):
        """Returns True if there are any quarantined objects in the Quarantine"""
        return self.nodes or self.edges or self.events



from __future__ import annotations

import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Protocol,
    Tuple,
    TypeVar,
    Union,
)

from multimethod import multimethod
from pydantic import BaseModel


class SpecProto(Protocol):
    def spec(self) -> Any:
        pass


T = TypeVar("T")


def unpack_object(obj: Union[Dict, BaseModel]) -> Dict:
    """

    Args:
        obj: The object to unpack, generally a dict or BaseModel

    Returns:

    Raises:

    """
    if isinstance(obj, Dict):
        return obj
    elif isinstance(obj, BaseModel):
        return obj.dict()
    else:
        raise NotImplementedError(f"No method to unpack objects of type {type(obj)}")


atomic = Union[int, float, complex, str, bool, uuid.UUID]


@multimethod
def merge(a, b):
    """The base merge function

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:

    Raises:

    """
    raise Exception(f"No merge logic for type {type(a)} and {type(b)}")


@merge.register
def merge_atomic(a: Any, b: T) -> T:
    """Merge an atomic value with any other value

    This function effectively handles cases where a value is being replaced.
    For example, if the value integer `5` needed to be updated to `6`, this function would handle that.

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        The new value

    Raises:

    """
    return b


@merge.register
def merge_missing(a: T, b: None) -> T:
    """Merge an object with a missing value

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        The original object

    Raises:

    """
    return a


@merge.register
def merge_dict_item(a: Dict, b: Dict) -> Dict:
    """Merge two dictionaries

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        The merged dictionary

    Raises:

    """
    result = {**a, **b}
    result.update({key: merge(a[key], b[key]) for key in a.keys() & b.keys()})
    return result


@merge.register
def merge_list(a: list, b: list) -> list:
    """Merge two lists

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        A merged list

    Raises:

    """
    return [*a, *b]


@merge.register
def merge_tuple(a: tuple, b: tuple) -> tuple:
    """Merge two tuples

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        A merged tuple

    Raises:

    """
    return *a, *b


@merge.register
def merge_set(a: set, b: set) -> set:
    """Merge two sets

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        A merged set

    Raises:

    """
    return b | a


@merge.register
def merge_pydantic(a: BaseModel, b: Any) -> BaseModel:
    """Merge a non-pydantic object into a pydantic model

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        An updated pydantic model

    Raises:

    """
    merged = merge(dict(a), b)
    return type(a)(**merged)


@merge.register
def merge_pydantic_right(a: T, b: BaseModel) -> T:
    """Merge a pydantic model into a non-pydantic model

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        An updated non-pydantic model

    Raises:

    """
    return merge(a, dict(b))


def merge_models(a: T, b: T) -> T:
    """This function is deprecated. Use `merge` instead

    Args:
        a: The first object to merge
        b: The second object to merge

    Returns:
        An updated model

    Raises:

    """
    a_type = type(a)
    merged = merge(a, b)
    return a_type(**merged)


def compute_graph_changes(
    items: List[SpecProto], active_items: List[SpecProto]
) -> Tuple[List[SpecProto], List[SpecProto], List[SpecProto]]:
    """Computes a graph update for a list of items and the corresponding set of currently active items.

    Args:
        items: The new list of graph nodes
        active_items: The current list of graph nodes. This does not have to be the full graph, only those relevant to the items set. Most commonly, these are drawn from the same data source.

    Returns:
        A three tuple of new items, updated items, and deleted items.
    Raises:

    """

    active_item_map = {hash(item.spec): item for item in active_items}
    item_map: Dict[int, SpecProto] = {hash(item.spec): item for item in items}

    new_item_keys = item_map.keys() - active_item_map.keys()
    deleted_source_item_keys = active_item_map.keys() - item_map.keys()
    updated_item_keys = item_map.keys() - new_item_keys

    deleted_from_sources = [active_item_map[k] for k in deleted_source_item_keys]

    new_items: List[SpecProto] = [item_map[k] for k in new_item_keys]
    updated_items = [item_map[k] for k in updated_item_keys if item_map[k] != active_item_map[k]]

    return new_items, updated_items, deleted_from_sources



from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from grai_schemas.generics import GraiBaseModel, MalformedMetadata
from grai_schemas.v1.generics import Code
from grai_schemas.v1.metadata.generics import GenericAttributes


class EdgeMetadataTypeLabels(Enum):
    """ """

    generic = "Generic"
    table_to_column = "TableToColumn"
    column_to_column = "ColumnToColumn"
    table_to_table = "TableToTable"


EdgeTypeLabelLiterals = Literal["Generic", "TableToColumn", "ColumnToColumn", "TableToTable"]


class BaseEdgeMetadataV1(GraiBaseModel):
    """Class definition of BaseEdgeMetadataV1

    Attributes:
        type: Object type of the Metadata e.g. NodeV1, EdgeV1, etc.
        version: Schema version of the metadata
        edge_type: The type of edge e.g. TableToColumn, ColumnToColumn, etc.
        edge_attributes: Attributes specific to the edge type
        tags: Tags associated with the edge

    """

    type: Literal["EdgeV1"] = "EdgeV1"
    version: Literal["v1"] = "v1"
    edge_type: EdgeTypeLabelLiterals
    edge_attributes: GenericAttributes
    tags: Optional[List[str]]
    code: Code = Code()


class MalformedEdgeMetadataV1(MalformedMetadata, BaseEdgeMetadataV1):
    """Class definition of MalformedEdgeMetadataV1

    Attributes:
        edge_type: The literal "Malformed"
        edge_attributes: Attributes specific to the edge type

    """

    edge_type: Optional[str] = "Malformed"  # type: ignore
    edge_attributes: Optional[Any] = GenericAttributes()  # type: ignore


class GenericEdgeMetadataV1(BaseEdgeMetadataV1):
    """Class definition of GenericEdgeMetadataV1

    Attributes:
        edge_type: The literal "Generic"
        edge_attributes: Attributes specific to the edge type

    """

    edge_type: Literal["Generic"]
    edge_attributes: GenericAttributes = GenericAttributes()


class TableToColumnAttributes(GenericAttributes):
    """Class definition of TableToColumnAttributes

    Attributes:
        version: Schema version of the attributes

    """

    version: Literal["v1"] = "v1"


class TableToColumnMetadata(BaseEdgeMetadataV1):
    """Class definition of TableToColumnMetadata

    Attributes:
        edge_type: The literal "TableToColumn"
        edge_attributes: Attributes specific to the edge type

    """

    edge_type: Literal["TableToColumn"]
    edge_attributes: TableToColumnAttributes = TableToColumnAttributes()


class TableToTableAttributes(GenericAttributes):
    """Class definition of TableToTableAttributes

    Attributes:
        version: Schema version of the attributes

    """

    version: Literal["v1"] = "v1"


class TableToTableMetadata(BaseEdgeMetadataV1):
    """Class definition of TableToTableMetadata

    Attributes:
        edge_type: The literal "TableToTable"
        edge_attributes: Attributes specific to the edge type

    """

    edge_type: Literal["TableToTable"]
    edge_attributes: TableToTableAttributes = TableToTableAttributes()


class ColumnToColumnAttributes(GenericAttributes):
    """Class definition of ColumnToColumnAttributes

    Attributes:
        version: Schema version of the attributes
        preserves_data_type: Whether the data type is conserved between the source and destination columns
        preserves_nullable: Whether the nullability is conserved between the source and destination columns
        preserves_unique: Whether uniqueness is conserved between the source and destination columns

    """

    version: Literal["v1"] = "v1"
    preserves_data_type: Optional[bool] = None
    preserves_nullable: Optional[bool] = None
    preserves_unique: Optional[bool] = None


class ColumnToColumnMetadata(BaseEdgeMetadataV1):
    """Class definition of ColumnToColumnMetadata

    Attributes:
        edge_type: The literal "ColumnToColumn"
        edge_attributes: Attributes specific to the edge type

    """

    edge_type: Literal["ColumnToColumn"]
    edge_attributes: ColumnToColumnAttributes = ColumnToColumnAttributes()


Metadata = Union[TableToColumnMetadata, TableToTableMetadata, ColumnToColumnMetadata, GenericEdgeMetadataV1]



import random

NOUNS = [
    "people",
    "history",
    "way",
    "art",
    "world",
    "information",
    "map",
    "two",
    "family",
    "government",
    "health",
    "system",
    "computer",
    "meat",
    "year",
    "thanks",
    "music",
    "person",
    "reading",
    "method",
    "data",
    "food",
    "understanding",
    "theory",
    "law",
    "bird",
    "literature",
    "problem",
    "software",
    "control",
    "knowledge",
    "power",
    "ability",
    "economics",
    "love",
    "internet",
    "television",
    "science",
    "library",
    "nature",
    "fact",
    "product",
    "idea",
    "temperature",
    "investment",
    "area",
    "society",
    "activity",
    "story",
    "industry",
    "media",
    "thing",
    "oven",
    "community",
    "definition",
    "safety",
    "quality",
    "development",
    "language",
    "management",
    "player",
    "variety",
    "video",
    "week",
    "security",
    "country",
    "exam",
    "movie",
    "organization",
    "equipment",
    "physics",
    "analysis",
    "policy",
    "series",
    "thought",
    "basis",
    "boyfriend",
    "direction",
    "strategy",
    "technology",
    "army",
    "camera",
    "freedom",
    "paper",
    "environment",
    "child",
    "instance",
    "month",
    "truth",
    "marketing",
    "university",
    "writing",
    "article",
    "department",
    "difference",
    "goal",
    "news",
    "audience",
    "fishing",
    "growth",
    "income",
    "marriage",
    "user",
    "combination",
    "failure",
    "meaning",
    "medicine",
    "philosophy",
    "teacher",
    "communication",
    "night",
    "chemistry",
    "disease",
    "disk",
    "energy",
    "nation",
    "road",
    "role",
    "soup",
    "advertising",
    "location",
    "success",
    "addition",
    "apartment",
    "education",
    "math",
    "moment",
    "painting",
    "politics",
    "attention",
    "decision",
    "event",
    "property",
    "shopping",
    "student",
    "wood",
    "competition",
    "distribution",
    "entertainment",
    "office",
    "population",
    "president",
    "unit",
    "category",
    "cigarette",
    "context",
    "introduction",
    "opportunity",
    "performance",
    "driver",
    "flight",
    "length",
    "magazine",
    "newspaper",
    "relationship",
    "teaching",
    "cell",
    "dealer",
    "finding",
    "lake",
    "member",
    "message",
    "phone",
    "scene",
    "appearance",
    "association",
    "concept",
    "customer",
    "death",
    "discussion",
    "housing",
    "inflation",
    "insurance",
    "mood",
    "woman",
    "advice",
    "blood",
    "effort",
    "expression",
    "importance",
    "opinion",
    "payment",
    "reality",
    "responsibility",
    "situation",
    "skill",
    "statement",
    "wealth",
    "application",
    "city",
    "county",
    "depth",
    "estate",
    "foundation",
    "grandmother",
    "heart",
    "perspective",
    "photo",
    "recipe",
    "studio",
    "topic",
    "collection",
    "depression",
    "imagination",
    "passion",
    "percentage",
    "resource",
    "setting",
    "ad",
    "agency",
    "college",
    "connection",
    "criticism",
    "debt",
    "description",
    "memory",
    "patience",
    "secretary",
    "solution",
    "administration",
    "aspect",
    "attitude",
    "director",
    "personality",
    "psychology",
    "recommendation",
    "response",
    "selection",
    "storage",
    "version",
    "alcohol",
    "argument",
    "complaint",
    "contract",
    "emphasis",
    "highway",
    "loss",
    "membership",
    "possession",
    "preparation",
    "steak",
    "union",
    "agreement",
    "cancer",
    "currency",
    "employment",
    "engineering",
    "entry",
    "interaction",
    "mixture",
    "preference",
    "region",
    "republic",
    "tradition",
    "virus",
    "actor",
    "classroom",
    "delivery",
    "device",
    "difficulty",
    "drama",
    "election",
    "engine",
    "football",
    "guidance",
    "hotel",
    "owner",
    "priority",
    "protection",
    "suggestion",
    "tension",
    "variation",
    "anxiety",
    "atmosphere",
    "awareness",
    "bath",
    "bread",
    "candidate",
    "climate",
    "comparison",
    "confusion",
    "construction",
    "elevator",
    "emotion",
    "employee",
    "employer",
    "guest",
    "height",
    "leadership",
    "mall",
    "manager",
    "operation",
    "recording",
    "sample",
    "transportation",
    "charity",
    "cousin",
    "disaster",
    "editor",
    "efficiency",
    "excitement",
    "extent",
    "feedback",
    "guitar",
    "homework",
    "leader",
    "mom",
    "outcome",
    "permission",
    "presentation",
    "promotion",
    "reflection",
    "refrigerator",
    "resolution",
    "revenue",
    "session",
    "singer",
    "tennis",
    "basket",
    "bonus",
    "cabinet",
    "childhood",
    "church",
    "clothes",
    "coffee",
    "dinner",
    "drawing",
    "hair",
    "hearing",
    "initiative",
    "judgment",
    "lab",
    "measurement",
    "mode",
    "mud",
    "orange",
    "poetry",
    "police",
    "possibility",
    "procedure",
    "queen",
    "ratio",
    "relation",
    "restaurant",
    "satisfaction",
    "sector",
    "signature",
    "significance",
    "song",
    "tooth",
    "town",
    "vehicle",
    "volume",
    "wife",
    "accident",
    "airport",
    "appointment",
    "arrival",
    "assumption",
    "baseball",
    "chapter",
    "committee",
    "conversation",
    "database",
    "enthusiasm",
    "error",
    "explanation",
    "farmer",
    "gate",
    "girl",
    "hall",
    "historian",
    "hospital",
    "injury",
    "instruction",
    "maintenance",
    "manufacturer",
    "meal",
    "perception",
    "pie",
    "poem",
    "presence",
    "proposal",
    "reception",
    "replacement",
    "revolution",
    "river",
    "son",
    "speech",
    "tea",
    "village",
    "warning",
    "winner",
    "worker",
    "writer",
    "assistance",
    "breath",
    "buyer",
    "chest",
    "chocolate",
    "conclusion",
    "contribution",
    "cookie",
    "courage",
    "dad",
    "desk",
    "drawer",
    "establishment",
    "examination",
    "garbage",
    "grocery",
    "honey",
    "impression",
    "improvement",
    "independence",
    "insect",
    "inspection",
    "inspector",
    "king",
    "ladder",
    "menu",
    "penalty",
    "piano",
    "potato",
    "profession",
    "professor",
    "quantity",
    "reaction",
    "requirement",
    "salad",
    "sister",
    "supermarket",
    "tongue",
    "weakness",
    "wedding",
    "affair",
    "ambition",
    "analyst",
    "apple",
    "assignment",
    "assistant",
    "bathroom",
    "bedroom",
    "beer",
    "birthday",
    "celebration",
    "championship",
    "cheek",
    "client",
    "consequence",
    "departure",
    "diamond",
    "dirt",
    "ear",
    "fortune",
    "friendship",
    "funeral",
    "gene",
    "girlfriend",
    "hat",
    "indication",
    "intention",
    "lady",
    "midnight",
    "negotiation",
    "obligation",
    "passenger",
    "pizza",
    "platform",
    "poet",
    "pollution",
    "recognition",
    "reputation",
    "shirt",
    "sir",
    "speaker",
    "stranger",
    "surgery",
    "sympathy",
    "tale",
    "throat",
    "trainer",
    "uncle",
    "youth",
    "time",
    "work",
    "film",
    "water",
    "money",
    "example",
    "while",
    "business",
    "study",
    "game",
    "life",
    "form",
    "air",
    "day",
    "place",
    "number",
    "part",
    "field",
    "fish",
    "back",
    "process",
    "heat",
    "hand",
    "experience",
    "job",
    "book",
    "end",
    "point",
    "type",
    "home",
    "economy",
    "value",
    "body",
    "market",
    "guide",
    "interest",
    "state",
    "radio",
    "course",
    "company",
    "price",
    "size",
    "card",
    "list",
    "mind",
    "trade",
    "line",
    "care",
    "group",
    "risk",
    "word",
    "fat",
    "force",
    "key",
    "light",
    "training",
    "name",
    "school",
    "top",
    "amount",
    "level",
    "order",
    "practice",
    "research",
    "sense",
    "service",
    "piece",
    "web",
    "boss",
    "sport",
    "fun",
    "house",
    "page",
    "term",
    "test",
    "answer",
    "sound",
    "focus",
    "matter",
    "kind",
    "soil",
    "board",
    "oil",
    "picture",
    "access",
    "garden",
    "range",
    "rate",
    "reason",
    "future",
    "site",
    "demand",
    "exercise",
    "image",
    "case",
    "cause",
    "coast",
    "action",
    "age",
    "bad",
    "boat",
    "record",
    "result",
    "section",
    "building",
    "mouse",
    "cash",
    "class",
    "nothing",
    "period",
    "plan",
    "store",
    "tax",
    "side",
    "subject",
    "space",
    "rule",
    "stock",
    "weather",
    "chance",
    "figure",
    "man",
    "model",
    "source",
    "beginning",
    "earth",
    "program",
    "chicken",
    "design",
    "feature",
    "head",
    "material",
    "purpose",
    "question",
    "rock",
    "salt",
    "act",
    "birth",
    "car",
    "dog",
    "object",
    "scale",
    "sun",
    "note",
    "profit",
    "rent",
    "speed",
    "style",
    "war",
    "bank",
    "craft",
    "half",
    "inside",
    "outside",
    "standard",
    "bus",
    "exchange",
    "eye",
    "fire",
    "position",
    "pressure",
    "stress",
    "advantage",
    "benefit",
    "box",
    "frame",
    "issue",
    "step",
    "cycle",
    "face",
    "item",
    "metal",
    "paint",
    "review",
    "room",
    "screen",
    "structure",
    "view",
    "account",
    "ball",
    "discipline",
    "medium",
    "share",
    "balance",
    "bit",
    "black",
    "bottom",
    "choice",
    "gift",
    "impact",
    "machine",
    "shape",
    "tool",
    "wind",
    "address",
    "average",
    "career",
    "culture",
    "morning",
    "pot",
    "sign",
    "table",
    "task",
    "condition",
    "contact",
    "credit",
    "egg",
    "hope",
    "ice",
    "network",
    "north",
    "square",
    "attempt",
    "date",
    "effect",
    "link",
    "post",
    "star",
    "voice",
    "capital",
    "challenge",
    "friend",
    "self",
    "shot",
    "brush",
    "couple",
    "debate",
    "exit",
    "front",
    "function",
    "lack",
    "living",
    "plant",
    "plastic",
    "spot",
    "summer",
    "taste",
    "theme",
    "track",
    "wing",
    "brain",
    "button",
    "click",
    "desire",
    "foot",
    "gas",
    "influence",
    "notice",
    "rain",
    "wall",
    "base",
    "damage",
    "distance",
    "feeling",
    "pair",
    "savings",
    "staff",
    "sugar",
    "target",
    "text",
    "animal",
    "author",
    "budget",
    "discount",
    "file",
    "ground",
    "lesson",
    "minute",
    "officer",
    "phase",
    "reference",
    "register",
    "sky",
    "stage",
    "stick",
    "title",
    "trouble",
    "bowl",
    "bridge",
    "campaign",
    "character",
    "club",
    "edge",
    "evidence",
    "fan",
    "letter",
    "lock",
    "maximum",
    "novel",
    "option",
    "pack",
    "park",
    "plenty",
    "quarter",
    "skin",
    "sort",
    "weight",
    "baby",
    "background",
    "carry",
    "dish",
    "factor",
    "fruit",
    "glass",
    "joint",
    "master",
    "muscle",
    "red",
    "strength",
    "traffic",
    "trip",
    "vegetable",
    "appeal",
    "chart",
    "gear",
    "ideal",
    "kitchen",
    "land",
    "log",
    "mother",
    "net",
    "party",
    "principle",
    "relative",
    "sale",
    "season",
    "signal",
    "spirit",
    "street",
    "tree",
    "wave",
    "belt",
    "bench",
    "commission",
    "copy",
    "drop",
    "minimum",
    "path",
    "progress",
    "project",
    "sea",
    "south",
    "status",
    "stuff",
    "ticket",
    "tour",
    "angle",
    "blue",
    "breakfast",
    "confidence",
    "daughter",
    "degree",
    "doctor",
    "dot",
    "dream",
    "duty",
    "essay",
    "father",
    "fee",
    "finance",
    "hour",
    "juice",
    "limit",
    "luck",
    "milk",
    "mouth",
    "peace",
    "pipe",
    "seat",
    "stable",
    "storm",
    "substance",
    "team",
    "trick",
    "afternoon",
    "bat",
    "beach",
    "blank",
    "catch",
    "chain",
    "consideration",
    "cream",
    "crew",
    "detail",
    "gold",
    "interview",
    "kid",
    "mark",
    "match",
    "mission",
    "pain",
    "pleasure",
    "score",
    "screw",
    "shop",
    "shower",
    "suit",
    "tone",
    "window",
    "agent",
    "band",
    "block",
    "bone",
    "calendar",
    "cap",
    "coat",
    "contest",
    "corner",
    "court",
    "cup",
    "district",
    "door",
    "east",
    "finger",
    "garage",
    "guarantee",
    "hole",
    "hook",
    "implement",
    "layer",
    "lecture",
    "lie",
    "manner",
    "meeting",
    "nose",
    "parking",
    "partner",
    "profile",
    "respect",
    "rice",
    "routine",
    "schedule",
    "swimming",
    "telephone",
    "tip",
    "winter",
    "airline",
    "bag",
    "battle",
    "bed",
    "bill",
    "bother",
    "cake",
    "code",
    "curve",
    "designer",
    "dimension",
    "dress",
    "ease",
    "emergency",
    "evening",
    "extension",
    "farm",
    "fight",
    "gap",
    "grade",
    "holiday",
    "horror",
    "horse",
    "host",
    "husband",
    "loan",
    "mistake",
    "mountain",
    "nail",
    "noise",
    "occasion",
    "package",
    "patient",
    "pause",
    "phrase",
    "proof",
    "race",
    "relief",
    "sand",
    "sentence",
    "shoulder",
    "smoke",
    "stomach",
    "string",
    "tourist",
    "towel",
    "vacation",
    "west",
    "wheel",
    "wine",
    "arm",
    "aside",
    "associate",
    "bet",
    "blow",
    "border",
    "branch",
    "breast",
    "brother",
    "buddy",
    "bunch",
    "chip",
    "coach",
    "cross",
    "document",
    "draft",
    "dust",
    "expert",
    "floor",
    "god",
    "golf",
    "habit",
    "iron",
    "judge",
    "knife",
    "landscape",
    "league",
    "mail",
    "mess",
    "native",
    "opening",
    "parent",
    "pattern",
    "pin",
    "pool",
    "pound",
    "request",
    "salary",
    "shame",
    "shelter",
    "shoe",
    "silver",
    "tackle",
    "tank",
    "trust",
    "assist",
    "bake",
    "bar",
    "bell",
    "bike",
    "blame",
    "boy",
    "brick",
    "chair",
    "closet",
    "clue",
    "collar",
    "comment",
    "conference",
    "devil",
    "diet",
    "fear",
    "fuel",
    "glove",
    "jacket",
    "lunch",
    "monitor",
    "mortgage",
    "nurse",
    "pace",
    "panic",
    "peak",
    "plane",
    "reward",
    "row",
    "sandwich",
    "shock",
    "spite",
    "spray",
    "surprise",
    "till",
    "transition",
    "weekend",
    "welcome",
    "yard",
    "alarm",
    "bend",
    "bicycle",
    "bite",
    "blind",
    "bottle",
    "cable",
    "candle",
    "clerk",
    "cloud",
    "concert",
    "counter",
    "flower",
    "grandfather",
    "harm",
    "knee",
    "lawyer",
    "leather",
    "load",
    "mirror",
    "neck",
    "pension",
    "plate",
    "purple",
    "ruin",
    "ship",
    "skirt",
    "slice",
    "snow",
    "specialist",
    "stroke",
    "switch",
    "trash",
    "tune",
    "zone",
    "anger",
    "award",
    "bid",
    "bitter",
    "boot",
    "bug",
    "camp",
    "candy",
    "carpet",
    "cat",
    "champion",
    "channel",
    "clock",
    "comfort",
    "cow",
    "crack",
    "engineer",
    "entrance",
    "fault",
    "grass",
    "guy",
    "hell",
    "highlight",
    "incident",
    "island",
    "joke",
    "jury",
    "leg",
    "lip",
    "mate",
    "motor",
    "nerve",
    "passage",
    "pen",
    "pride",
    "priest",
    "prize",
    "promise",
    "resident",
    "resort",
    "ring",
    "roof",
    "rope",
    "sail",
    "scheme",
    "script",
    "sock",
    "station",
    "toe",
    "tower",
    "truck",
    "witness",
    "a",
    "you",
    "it",
    "can",
    "will",
    "if",
    "one",
    "many",
    "most",
    "other",
    "use",
    "make",
    "good",
    "look",
    "help",
    "go",
    "great",
    "being",
    "few",
    "might",
    "still",
    "public",
    "read",
    "keep",
    "start",
    "give",
    "human",
    "local",
    "general",
    "she",
    "specific",
    "long",
    "play",
    "feel",
    "high",
    "tonight",
    "put",
    "common",
    "set",
    "change",
    "simple",
    "past",
    "big",
    "possible",
    "particular",
    "today",
    "major",
    "personal",
    "current",
    "national",
    "cut",
    "natural",
    "physical",
    "show",
    "try",
    "check",
    "second",
    "call",
    "move",
    "pay",
    "let",
    "increase",
    "single",
    "individual",
    "turn",
    "ask",
    "buy",
    "guard",
    "hold",
    "main",
    "offer",
    "potential",
    "professional",
    "international",
    "travel",
    "cook",
    "alternative",
    "following",
    "special",
    "working",
    "whole",
    "dance",
    "excuse",
    "cold",
    "commercial",
    "low",
    "purchase",
    "deal",
    "primary",
    "worth",
    "fall",
    "necessary",
    "positive",
    "produce",
    "search",
    "present",
    "spend",
    "talk",
    "creative",
    "tell",
    "cost",
    "drive",
    "green",
    "support",
    "glad",
    "remove",
    "return",
    "run",
    "complex",
    "due",
    "effective",
    "middle",
    "regular",
    "reserve",
    "independent",
    "leave",
    "original",
    "reach",
    "rest",
    "serve",
    "watch",
    "beautiful",
    "charge",
    "active",
    "break",
    "negative",
    "safe",
    "stay",
    "visit",
    "visual",
    "affect",
    "cover",
    "report",
    "rise",
    "walk",
    "white",
    "beyond",
    "junior",
    "pick",
    "unique",
    "anything",
    "classic",
    "final",
    "lift",
    "mix",
    "private",
    "stop",
    "teach",
    "western",
    "concern",
    "familiar",
    "fly",
    "official",
    "broad",
    "comfortable",
    "gain",
    "maybe",
    "rich",
    "save",
    "stand",
    "young",
    "fail",
    "heavy",
    "hello",
    "lead",
    "listen",
    "valuable",
    "worry",
    "handle",
    "leading",
    "meet",
    "release",
    "sell",
    "finish",
    "normal",
    "press",
    "ride",
    "secret",
    "spread",
    "spring",
    "tough",
    "wait",
    "brown",
    "deep",
    "display",
    "flow",
    "hit",
    "objective",
    "shoot",
    "touch",
    "cancel",
    "chemical",
    "cry",
    "dump",
    "extreme",
    "push",
    "conflict",
    "eat",
    "fill",
    "formal",
    "jump",
    "kick",
    "opposite",
    "pass",
    "pitch",
    "remote",
    "total",
    "treat",
    "vast",
    "abuse",
    "beat",
    "burn",
    "deposit",
    "print",
    "raise",
    "sleep",
    "somewhere",
    "advance",
    "anywhere",
    "consist",
    "dark",
    "double",
    "draw",
    "equal",
    "fix",
    "hire",
    "internal",
    "join",
    "kill",
    "sensitive",
    "tap",
    "win",
    "attack",
    "claim",
    "constant",
    "drag",
    "drink",
    "guess",
    "minor",
    "pull",
    "raw",
    "soft",
    "solid",
    "wear",
    "weird",
    "wonder",
    "annual",
    "count",
    "dead",
    "doubt",
    "feed",
    "forever",
    "impress",
    "nobody",
    "repeat",
    "round",
    "sing",
    "slide",
    "strip",
    "whereas",
    "wish",
    "combine",
    "command",
    "dig",
    "divide",
    "equivalent",
    "hang",
    "hunt",
    "initial",
    "march",
    "mention",
    "smell",
    "spiritual",
    "survey",
    "tie",
    "adult",
    "brief",
    "crazy",
    "escape",
    "gather",
    "hate",
    "prior",
    "repair",
    "rough",
    "sad",
    "scratch",
    "sick",
    "strike",
    "employ",
    "external",
    "hurt",
    "illegal",
    "laugh",
    "lay",
    "mobile",
    "nasty",
    "ordinary",
    "respond",
    "royal",
    "senior",
    "split",
    "strain",
    "struggle",
    "swim",
    "train",
    "upper",
    "wash",
    "yellow",
    "convert",
    "crash",
    "dependent",
    "fold",
    "funny",
    "grab",
    "hide",
    "miss",
    "permit",
    "quote",
    "recover",
    "resolve",
    "roll",
    "sink",
    "slip",
    "spare",
    "suspect",
    "sweet",
    "swing",
    "twist",
    "upstairs",
    "usual",
    "abroad",
    "brave",
    "calm",
    "concentrate",
    "estimate",
    "grand",
    "male",
    "mine",
    "prompt",
    "quiet",
    "refuse",
    "regret",
    "reveal",
    "rush",
    "shake",
    "shift",
    "shine",
    "steal",
    "suck",
    "surround",
    "anybody",
    "bear",
    "brilliant",
    "dare",
    "dear",
    "delay",
    "drunk",
    "female",
    "hurry",
    "inevitable",
    "invite",
    "kiss",
    "neat",
    "pop",
    "punch",
    "quit",
    "reply",
    "representative",
    "resist",
    "rip",
    "rub",
    "silly",
    "smile",
    "spell",
    "stretch",
    "stupid",
    "tear",
    "temporary",
    "tomorrow",
    "wake",
    "wrap",
    "yesterday",
]

ADJECTIVES = [
    "abandoned",
    "able",
    "absolute",
    "adorable",
    "adventurous",
    "academic",
    "acceptable",
    "acclaimed",
    "accomplished",
    "accurate",
    "aching",
    "acidic",
    "acrobatic",
    "active",
    "actual",
    "adept",
    "admirable",
    "admired",
    "adolescent",
    "adorable",
    "adored",
    "advanced",
    "afraid",
    "affectionate",
    "aged",
    "aggravating",
    "aggressive",
    "agile",
    "agitated",
    "agonizing",
    "agreeable",
    "ajar",
    "alarmed",
    "alarming",
    "alert",
    "alienated",
    "alive",
    "all",
    "altruistic",
    "amazing",
    "ambitious",
    "ample",
    "amused",
    "amusing",
    "anchored",
    "ancient",
    "angelic",
    "angry",
    "anguished",
    "animated",
    "annual",
    "another",
    "antique",
    "anxious",
    "any",
    "apprehensive",
    "appropriate",
    "apt",
    "arctic",
    "arid",
    "aromatic",
    "artistic",
    "ashamed",
    "assured",
    "astonishing",
    "athletic",
    "attached",
    "attentive",
    "attractive",
    "austere",
    "authentic",
    "authorized",
    "automatic",
    "avaricious",
    "average",
    "aware",
    "awesome",
    "awful",
    "awkward",
    "babyish",
    "bad",
    "back",
    "baggy",
    "bare",
    "barren",
    "basic",
    "beautiful",
    "belated",
    "beloved",
    "beneficial",
    "better",
    "best",
    "bewitched",
    "big",
    "big-hearted",
    "biodegradable",
    "bite-sized",
    "bitter",
    "black",
    "black-and-white",
    "bland",
    "blank",
    "blaring",
    "bleak",
    "blind",
    "blissful",
    "blond",
    "blue",
    "blushing",
    "bogus",
    "boiling",
    "bold",
    "bony",
    "boring",
    "bossy",
    "both",
    "bouncy",
    "bountiful",
    "bowed",
    "brave",
    "breakable",
    "brief",
    "bright",
    "brilliant",
    "brisk",
    "broken",
    "bronze",
    "brown",
    "bruised",
    "bubbly",
    "bulky",
    "bumpy",
    "buoyant",
    "burdensome",
    "burly",
    "bustling",
    "busy",
    "buttery",
    "buzzing",
    "calculating",
    "calm",
    "candid",
    "canine",
    "capital",
    "carefree",
    "careful",
    "careless",
    "caring",
    "cautious",
    "cavernous",
    "celebrated",
    "charming",
    "cheap",
    "cheerful",
    "cheery",
    "chief",
    "chilly",
    "chubby",
    "circular",
    "classic",
    "clean",
    "clear",
    "clear-cut",
    "clever",
    "close",
    "closed",
    "cloudy",
    "clueless",
    "clumsy",
    "cluttered",
    "coarse",
    "cold",
    "colorful",
    "colorless",
    "colossal",
    "comfortable",
    "common",
    "compassionate",
    "competent",
    "complete",
    "complex",
    "complicated",
    "composed",
    "concerned",
    "concrete",
    "confused",
    "conscious",
    "considerate",
    "constant",
    "content",
    "conventional",
    "cooked",
    "cool",
    "cooperative",
    "coordinated",
    "corny",
    "corrupt",
    "costly",
    "courageous",
    "courteous",
    "crafty",
    "crazy",
    "creamy",
    "creative",
    "creepy",
    "criminal",
    "crisp",
    "critical",
    "crooked",
    "crowded",
    "cruel",
    "crushing",
    "cuddly",
    "cultivated",
    "cultured",
    "cumbersome",
    "curly",
    "curvy",
    "cute",
    "cylindrical",
    "damaged",
    "damp",
    "dangerous",
    "dapper",
    "daring",
    "darling",
    "dark",
    "dazzling",
    "dead",
    "deadly",
    "deafening",
    "dear",
    "dearest",
    "decent",
    "decimal",
    "decisive",
    "deep",
    "defenseless",
    "defensive",
    "defiant",
    "deficient",
    "definite",
    "definitive",
    "delayed",
    "delectable",
    "delicious",
    "delightful",
    "delirious",
    "demanding",
    "dense",
    "dental",
    "dependable",
    "dependent",
    "descriptive",
    "deserted",
    "detailed",
    "determined",
    "devoted",
    "different",
    "difficult",
    "digital",
    "diligent",
    "dim",
    "dimpled",
    "dimwitted",
    "direct",
    "disastrous",
    "discrete",
    "disfigured",
    "disgusting",
    "disloyal",
    "dismal",
    "distant",
    "downright",
    "dreary",
    "dirty",
    "disguised",
    "dishonest",
    "dismal",
    "distant",
    "distinct",
    "distorted",
    "dizzy",
    "dopey",
    "doting",
    "double",
    "downright",
    "drab",
    "drafty",
    "dramatic",
    "dreary",
    "droopy",
    "dry",
    "dual",
    "dull",
    "dutiful",
    "each",
    "eager",
    "earnest",
    "early",
    "easy",
    "easy-going",
    "ecstatic",
    "edible",
    "educated",
    "elaborate",
    "elastic",
    "elated",
    "elderly",
    "electric",
    "elegant",
    "elementary",
    "elliptical",
    "embarrassed",
    "embellished",
    "eminent",
    "emotional",
    "empty",
    "enchanted",
    "enchanting",
    "energetic",
    "enlightened",
    "enormous",
    "enraged",
    "entire",
    "envious",
    "equal",
    "equatorial",
    "essential",
    "esteemed",
    "ethical",
    "euphoric",
    "even",
    "evergreen",
    "everlasting",
    "every",
    "evil",
    "exalted",
    "excellent",
    "exemplary",
    "exhausted",
    "excitable",
    "excited",
    "exciting",
    "exotic",
    "expensive",
    "experienced",
    "expert",
    "extraneous",
    "extroverted",
    "extra-large",
    "extra-small",
    "fabulous",
    "failing",
    "faint",
    "fair",
    "faithful",
    "fake",
    "false",
    "familiar",
    "famous",
    "fancy",
    "fantastic",
    "far",
    "faraway",
    "far-flung",
    "far-off",
    "fast",
    "fat",
    "fatal",
    "fatherly",
    "favorable",
    "favorite",
    "fearful",
    "fearless",
    "feisty",
    "feline",
    "female",
    "feminine",
    "few",
    "fickle",
    "filthy",
    "fine",
    "finished",
    "firm",
    "first",
    "firsthand",
    "fitting",
    "fixed",
    "flaky",
    "flamboyant",
    "flashy",
    "flat",
    "flawed",
    "flawless",
    "flickering",
    "flimsy",
    "flippant",
    "flowery",
    "fluffy",
    "fluid",
    "flustered",
    "focused",
    "fond",
    "foolhardy",
    "foolish",
    "forceful",
    "forked",
    "formal",
    "forsaken",
    "forthright",
    "fortunate",
    "fragrant",
    "frail",
    "frank",
    "frayed",
    "free",
    "French",
    "fresh",
    "frequent",
    "friendly",
    "frightened",
    "frightening",
    "frigid",
    "frilly",
    "frizzy",
    "frivolous",
    "front",
    "frosty",
    "frozen",
    "frugal",
    "fruitful",
    "full",
    "fumbling",
    "functional",
    "funny",
    "fussy",
    "fuzzy",
    "gargantuan",
    "gaseous",
    "general",
    "generous",
    "gentle",
    "genuine",
    "giant",
    "giddy",
    "gigantic",
    "gifted",
    "giving",
    "glamorous",
    "glaring",
    "glass",
    "gleaming",
    "gleeful",
    "glistening",
    "glittering",
    "gloomy",
    "glorious",
    "glossy",
    "glum",
    "golden",
    "good",
    "good-natured",
    "gorgeous",
    "graceful",
    "gracious",
    "grand",
    "grandiose",
    "granular",
    "grateful",
    "grave",
    "gray",
    "great",
    "greedy",
    "green",
    "gregarious",
    "grim",
    "grimy",
    "gripping",
    "grizzled",
    "gross",
    "grotesque",
    "grouchy",
    "grounded",
    "growing",
    "growling",
    "grown",
    "grubby",
    "gruesome",
    "grumpy",
    "guilty",
    "gullible",
    "gummy",
    "hairy",
    "half",
    "handmade",
    "handsome",
    "handy",
    "happy",
    "happy-go-lucky",
    "hard",
    "hard-to-find",
    "harmful",
    "harmless",
    "harmonious",
    "harsh",
    "hasty",
    "hateful",
    "haunting",
    "healthy",
    "heartfelt",
    "hearty",
    "heavenly",
    "heavy",
    "hefty",
    "helpful",
    "helpless",
    "hidden",
    "hideous",
    "high",
    "high-level",
    "hilarious",
    "hoarse",
    "hollow",
    "homely",
    "honest",
    "honorable",
    "honored",
    "hopeful",
    "horrible",
    "hospitable",
    "hot",
    "huge",
    "humble",
    "humiliating",
    "humming",
    "humongous",
    "hungry",
    "hurtful",
    "husky",
    "icky",
    "icy",
    "ideal",
    "idealistic",
    "identical",
    "idle",
    "idiotic",
    "idolized",
    "ignorant",
    "ill",
    "illegal",
    "ill-fated",
    "ill-informed",
    "illiterate",
    "illustrious",
    "imaginary",
    "imaginative",
    "immaculate",
    "immaterial",
    "immediate",
    "immense",
    "impassioned",
    "impeccable",
    "impartial",
    "imperfect",
    "imperturbable",
    "impish",
    "impolite",
    "important",
    "impossible",
    "impractical",
    "impressionable",
    "impressive",
    "improbable",
    "impure",
    "inborn",
    "incomparable",
    "incompatible",
    "incomplete",
    "inconsequential",
    "incredible",
    "indelible",
    "inexperienced",
    "indolent",
    "infamous",
    "infantile",
    "infatuated",
    "inferior",
    "infinite",
    "informal",
    "innocent",
    "insecure",
    "insidious",
    "insignificant",
    "insistent",
    "instructive",
    "insubstantial",
    "intelligent",
    "intent",
    "intentional",
    "interesting",
    "internal",
    "international",
    "intrepid",
    "ironclad",
    "irresponsible",
    "irritating",
    "itchy",
    "jaded",
    "jagged",
    "jam-packed",
    "jaunty",
    "jealous",
    "jittery",
    "joint",
    "jolly",
    "jovial",
    "joyful",
    "joyous",
    "jubilant",
    "judicious",
    "juicy",
    "jumbo",
    "junior",
    "jumpy",
    "juvenile",
    "kaleidoscopic",
    "keen",
    "key",
    "kind",
    "kindhearted",
    "kindly",
    "klutzy",
    "knobby",
    "knotty",
    "knowledgeable",
    "knowing",
    "known",
    "kooky",
    "kosher",
    "lame",
    "lanky",
    "large",
    "last",
    "lasting",
    "late",
    "lavish",
    "lawful",
    "lazy",
    "leading",
    "lean",
    "leafy",
    "left",
    "legal",
    "legitimate",
    "light",
    "lighthearted",
    "likable",
    "likely",
    "limited",
    "limp",
    "limping",
    "linear",
    "lined",
    "liquid",
    "little",
    "live",
    "lively",
    "livid",
    "loathsome",
    "lone",
    "lonely",
    "long",
    "long-term",
    "loose",
    "lopsided",
    "lost",
    "loud",
    "lovable",
    "lovely",
    "loving",
    "low",
    "loyal",
    "lucky",
    "lumbering",
    "luminous",
    "lumpy",
    "lustrous",
    "luxurious",
    "mad",
    "made-up",
    "magnificent",
    "majestic",
    "major",
    "male",
    "mammoth",
    "married",
    "marvelous",
    "masculine",
    "massive",
    "mature",
    "meager",
    "mealy",
    "mean",
    "measly",
    "meaty",
    "medical",
    "mediocre",
    "medium",
    "meek",
    "mellow",
    "melodic",
    "memorable",
    "menacing",
    "merry",
    "messy",
    "metallic",
    "mild",
    "milky",
    "mindless",
    "miniature",
    "minor",
    "minty",
    "miserable",
    "miserly",
    "misguided",
    "misty",
    "mixed",
    "modern",
    "modest",
    "moist",
    "monstrous",
    "monthly",
    "monumental",
    "moral",
    "mortified",
    "motherly",
    "motionless",
    "mountainous",
    "muddy",
    "muffled",
    "multicolored",
    "mundane",
    "murky",
    "mushy",
    "musty",
    "muted",
    "mysterious",
    "naive",
    "narrow",
    "nasty",
    "natural",
    "naughty",
    "nautical",
    "near",
    "neat",
    "necessary",
    "needy",
    "negative",
    "neglected",
    "negligible",
    "neighboring",
    "nervous",
    "new",
    "next",
    "nice",
    "nifty",
    "nimble",
    "nippy",
    "nocturnal",
    "noisy",
    "nonstop",
    "normal",
    "notable",
    "noted",
    "noteworthy",
    "novel",
    "noxious",
    "numb",
    "nutritious",
    "nutty",
    "obedient",
    "obese",
    "oblong",
    "oily",
    "oblong",
    "obvious",
    "occasional",
    "odd",
    "oddball",
    "offbeat",
    "offensive",
    "official",
    "old",
    "old-fashioned",
    "only",
    "open",
    "optimal",
    "optimistic",
    "opulent",
    "orange",
    "orderly",
    "organic",
    "ornate",
    "ornery",
    "ordinary",
    "original",
    "other",
    "our",
    "outlying",
    "outgoing",
    "outlandish",
    "outrageous",
    "outstanding",
    "oval",
    "overcooked",
    "overdue",
    "overjoyed",
    "overlooked",
    "palatable",
    "pale",
    "paltry",
    "parallel",
    "parched",
    "partial",
    "passionate",
    "past",
    "pastel",
    "peaceful",
    "peppery",
    "perfect",
    "perfumed",
    "periodic",
    "perky",
    "personal",
    "pertinent",
    "pesky",
    "pessimistic",
    "petty",
    "phony",
    "physical",
    "piercing",
    "pink",
    "pitiful",
    "plain",
    "plaintive",
    "plastic",
    "playful",
    "pleasant",
    "pleased",
    "pleasing",
    "plump",
    "plush",
    "polished",
    "polite",
    "political",
    "pointed",
    "pointless",
    "poised",
    "poor",
    "popular",
    "portly",
    "posh",
    "positive",
    "possible",
    "potable",
    "powerful",
    "powerless",
    "practical",
    "precious",
    "present",
    "prestigious",
    "pretty",
    "precious",
    "previous",
    "pricey",
    "prickly",
    "primary",
    "prime",
    "pristine",
    "private",
    "prize",
    "probable",
    "productive",
    "profitable",
    "profuse",
    "proper",
    "proud",
    "prudent",
    "punctual",
    "pungent",
    "puny",
    "pure",
    "purple",
    "pushy",
    "putrid",
    "puzzled",
    "puzzling",
    "quaint",
    "qualified",
    "quarrelsome",
    "quarterly",
    "queasy",
    "querulous",
    "questionable",
    "quick",
    "quick-witted",
    "quiet",
    "quintessential",
    "quirky",
    "quixotic",
    "quizzical",
    "radiant",
    "ragged",
    "rapid",
    "rare",
    "rash",
    "raw",
    "recent",
    "reckless",
    "rectangular",
    "ready",
    "real",
    "realistic",
    "reasonable",
    "red",
    "reflecting",
    "regal",
    "regular",
    "reliable",
    "relieved",
    "remarkable",
    "remorseful",
    "remote",
    "repentant",
    "required",
    "respectful",
    "responsible",
    "repulsive",
    "revolving",
    "rewarding",
    "rich",
    "rigid",
    "right",
    "ringed",
    "ripe",
    "roasted",
    "robust",
    "rosy",
    "rotating",
    "rotten",
    "rough",
    "round",
    "rowdy",
    "royal",
    "rubbery",
    "rundown",
    "ruddy",
    "rude",
    "runny",
    "rural",
    "rusty",
    "sad",
    "safe",
    "salty",
    "same",
    "sandy",
    "sane",
    "sarcastic",
    "sardonic",
    "satisfied",
    "scaly",
    "scarce",
    "scared",
    "scary",
    "scented",
    "scholarly",
    "scientific",
    "scornful",
    "scratchy",
    "scrawny",
    "second",
    "secondary",
    "second-hand",
    "secret",
    "self-assured",
    "self-reliant",
    "selfish",
    "sentimental",
    "separate",
    "serene",
    "serious",
    "serpentine",
    "several",
    "severe",
    "shabby",
    "shadowy",
    "shady",
    "shallow",
    "shameful",
    "shameless",
    "sharp",
    "shimmering",
    "shiny",
    "shocked",
    "shocking",
    "shoddy",
    "short",
    "short-term",
    "showy",
    "shrill",
    "shy",
    "sick",
    "silent",
    "silky",
    "silly",
    "silver",
    "similar",
    "simple",
    "simplistic",
    "sinful",
    "single",
    "sizzling",
    "skeletal",
    "skinny",
    "sleepy",
    "slight",
    "slim",
    "slimy",
    "slippery",
    "slow",
    "slushy",
    "small",
    "smart",
    "smoggy",
    "smooth",
    "smug",
    "snappy",
    "snarling",
    "sneaky",
    "sniveling",
    "snoopy",
    "sociable",
    "soft",
    "soggy",
    "solid",
    "somber",
    "some",
    "spherical",
    "sophisticated",
    "sore",
    "sorrowful",
    "soulful",
    "soupy",
    "sour",
    "Spanish",
    "sparkling",
    "sparse",
    "specific",
    "spectacular",
    "speedy",
    "spicy",
    "spiffy",
    "spirited",
    "spiteful",
    "splendid",
    "spotless",
    "spotted",
    "spry",
    "square",
    "squeaky",
    "squiggly",
    "stable",
    "staid",
    "stained",
    "stale",
    "standard",
    "starchy",
    "stark",
    "starry",
    "steep",
    "sticky",
    "stiff",
    "stimulating",
    "stingy",
    "stormy",
    "straight",
    "strange",
    "steel",
    "strict",
    "strident",
    "striking",
    "striped",
    "strong",
    "studious",
    "stunning",
    "stupendous",
    "stupid",
    "sturdy",
    "stylish",
    "subdued",
    "submissive",
    "substantial",
    "subtle",
    "suburban",
    "sudden",
    "sugary",
    "sunny",
    "super",
    "superb",
    "superficial",
    "superior",
    "supportive",
    "sure-footed",
    "surprised",
    "suspicious",
    "svelte",
    "sweaty",
    "sweet",
    "sweltering",
    "swift",
    "sympathetic",
    "tall",
    "talkative",
    "tame",
    "tan",
    "tangible",
    "tart",
    "tasty",
    "tattered",
    "taut",
    "tedious",
    "teeming",
    "tempting",
    "tender",
    "tense",
    "tepid",
    "terrible",
    "terrific",
    "testy",
    "thankful",
    "that",
    "these",
    "thick",
    "thin",
    "third",
    "thirsty",
    "this",
    "thorough",
    "thorny",
    "those",
    "thoughtful",
    "threadbare",
    "thrifty",
    "thunderous",
    "tidy",
    "tight",
    "timely",
    "tinted",
    "tiny",
    "tired",
    "torn",
    "total",
    "tough",
    "traumatic",
    "treasured",
    "tremendous",
    "tragic",
    "trained",
    "tremendous",
    "triangular",
    "tricky",
    "trifling",
    "trim",
    "trivial",
    "troubled",
    "true",
    "trusting",
    "trustworthy",
    "trusty",
    "truthful",
    "tubby",
    "turbulent",
    "twin",
    "ugly",
    "ultimate",
    "unacceptable",
    "unaware",
    "uncomfortable",
    "uncommon",
    "unconscious",
    "understated",
    "unequaled",
    "uneven",
    "unfinished",
    "unfit",
    "unfolded",
    "unfortunate",
    "unhappy",
    "unhealthy",
    "uniform",
    "unimportant",
    "unique",
    "united",
    "unkempt",
    "unknown",
    "unlawful",
    "unlined",
    "unlucky",
    "unnatural",
    "unpleasant",
    "unrealistic",
    "unripe",
    "unruly",
    "unselfish",
    "unsightly",
    "unsteady",
    "unsung",
    "untidy",
    "untimely",
    "untried",
    "untrue",
    "unused",
    "unusual",
    "unwelcome",
    "unwieldy",
    "unwilling",
    "unwitting",
    "unwritten",
    "upbeat",
    "upright",
    "upset",
    "urban",
    "usable",
    "used",
    "useful",
    "useless",
    "utilized",
    "utter",
    "vacant",
    "vague",
    "vain",
    "valid",
    "valuable",
    "vapid",
    "variable",
    "vast",
    "velvety",
    "venerated",
    "vengeful",
    "verifiable",
    "vibrant",
    "vicious",
    "victorious",
    "vigilant",
    "vigorous",
    "villainous",
    "violet",
    "violent",
    "virtual",
    "virtuous",
    "visible",
    "vital",
    "vivacious",
    "vivid",
    "voluminous",
    "wan",
    "warlike",
    "warm",
    "warmhearted",
    "warped",
    "wary",
    "wasteful",
    "watchful",
    "waterlogged",
    "watery",
    "wavy",
    "wealthy",
    "weak",
    "weary",
    "webbed",
    "wee",
    "weekly",
    "weepy",
    "weighty",
    "weird",
    "welcome",
    "well-documented",
    "well-groomed",
    "well-informed",
    "well-lit",
    "well-made",
    "well-off",
    "well-to-do",
    "well-worn",
    "wet",
    "which",
    "whimsical",
    "whirlwind",
    "whispered",
    "white",
    "whole",
    "whopping",
    "wicked",
    "wide",
    "wide-eyed",
    "wiggly",
    "wild",
    "willing",
    "wilted",
    "winding",
    "windy",
    "winged",
    "wiry",
    "wise",
    "witty",
    "wobbly",
    "woeful",
    "wonderful",
    "wooden",
    "woozy",
    "wordy",
    "worldly",
    "worn",
    "worried",
    "worrisome",
    "worse",
    "worst",
    "worthless",
    "worthwhile",
    "worthy",
    "wrathful",
    "wretched",
    "writhing",
    "wrong",
    "wry",
    "yawning",
    "yearly",
    "yellow",
    "yellowish",
    "young",
    "youthful",
    "yummy",
    "zany",
    "zealous",
    "zesty",
    "zigzag",
]


def get_human_id():
    """generates a human readable id following the pattern {adjective}_{noun}"""
    return f"{random.choice(ADJECTIVES)}_{random.choice(NOUNS)}"



from datetime import datetime
from typing import Dict, Literal, Union
from uuid import UUID

from grai_schemas.generics import GraiBaseModel
from grai_schemas.v1.workspace import WorkspaceSpec
from pydantic import Json


class EventSpec(GraiBaseModel):
    """Class definition of EventSpec

    Attributes:
        id: UUID of the event.
        connection_id: UUID of the events connection.
        date: Datetime of the event.
        workspace: WorkspaceSpec of the event.
        diff: Json of the changes from the event.

    """

    id: UUID
    connection_id: UUID
    date: datetime
    workspace: Union[str, WorkspaceSpec]
    diff: Json

    def __hash__(self) -> int:
        return hash(self.id)


class EventV1(GraiBaseModel):
    """Class definition of EventV1

    Attributes:
        type: Object type of the Metadata e.g. NodeV1, EdgeV1, etc.
        version: Schema version of the object.
        spec: The event specification.

    """

    type: Literal["Event"] = "Event"
    version: Literal["v1"] = "v1"
    spec: EventSpec

    @classmethod
    def from_spec(cls, spec_dict: Dict) -> "EventV1":
        """

        Args:
            spec_dict (Dict):

        Returns:
            An EventV1 instance
        Raises:

        """
        return cls(version="v1", type="Event", spec=spec_dict)

    def __hash__(self):
        return hash(self.spec)



from typing import Union

from grai_schemas import v1
from grai_schemas.generics import GraiBaseModel, MalformedMetadata
from grai_schemas.v1.metadata import edges as edge_v1
from grai_schemas.v1.metadata import nodes as node_v1


class GraiMetadata(GraiBaseModel):
    """Class definition of GraiMetadata

    Attributes:
        grai: Grai defined metadata attributes used to drive application logic.

    """

    grai: Union[node_v1.Metadata, edge_v1.Metadata]


Node = v1.node.NodeV1
Edge = v1.edge.EdgeV1


SourcedNode = v1.node.SourcedNodeV1
SourcedEdge = v1.edge.SourcedEdgeV1
Source = v1.source.SourceV1
Event = v1.events.EventV1
Workspace = v1.workspace.WorkspaceV1
Organisation = v1.organization.OrganisationV1

GraiType = Union[Node, Edge, SourcedNode, SourcedEdge, Source, Event, Workspace, Organisation]

__all__ = [
    "GraiMetadata",
    "Node",
    "Edge",
    "Event",
    "SourcedNode",
    "SourcedEdge",
    "Source",
    "Workspace",
    "Organisation",
    "GraiType",
]



import json
import os
import pathlib
from datetime import date, datetime, timezone
from enum import Enum
from json import JSONEncoder
from pathlib import Path
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Type,
    Union,
)
from uuid import UUID

import yaml
from pydantic import BaseModel
from pydantic.json import pydantic_encoder


def to_ecma262(dt: datetime) -> str:
    """Convert a datetime to a string in ECMA-262 format

    This function aims to maintain compatibility with Django default datetime behavior which itself follows the ECMA-262 standard.

    Args:
        dt: The datetime to convert

    Returns:
        The datetime in ECMA-262 format
    """

    dt_utc = dt.astimezone(timezone.utc)
    return dt_utc.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"


class GraiEncoder(JSONEncoder):
    """The default JSON encoder for Grai.

    This encoder provides default serialization for a variety of datatypes including:
    * Enums
    * Pydantic models
    * UUIDs
    * Paths
    * Datetimes & dates
    * Sets
    * etc...

    Which should be reused for compatibility purposes.

    """

    def default(self, obj: Any):
        if isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, BaseModel):
            return obj.dict()
        elif isinstance(obj, (UUID, pathlib.PosixPath, pathlib.WindowsPath)):
            return str(obj)
        elif isinstance(obj, datetime):
            return to_ecma262(obj)
        elif isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        elif isinstance(obj, set):
            return list(obj)

        return super().default(obj)


encoder = GraiEncoder()


def dump_json(v, *, default: Optional[Callable] = None) -> str:
    """Dump an object to JSON following Grai's serialization rules

    This uses the GraiEncoder to serialize objects to JSON.

    Args:
        v: The object to dump
        default: A default function to use for serialization if extra types are required.

    Returns:
        The JSON string

    """
    if default is None:
        return json.dumps(v, cls=GraiEncoder)
    else:
        return json.dumps(v, default=default, cls=GraiEncoder)


def load_json(v: str) -> Any:
    """Returns a JSON object from a string"""
    return json.loads(v)


class GraiYamlSerializer:
    """A YAML serializer for Grai

    The GraiYamlSerializer provides a simple interface for serializing and deserializing YAML files which complies with Grai's serialization rules.

    """

    @staticmethod
    def load(stream: Union[str, Path, IO]) -> Union[Dict, List[Dict]]:
        """

        Args:
            stream: The stream to load from. This can be a string, a Path, or a file-like object.

        Returns:
            Either a dictionary or a list of dictionaries depending on the input
        Raises:

        """
        if isinstance(stream, Path):
            with open(stream, "r") as file:
                result = list(yaml.safe_load_all(file))
        elif isinstance(stream, str) and (stream.endswith((".yml", ".yaml")) or os.path.exists(stream)):
            with open(stream, "r") as file:
                result = list(yaml.safe_load_all(file))
        else:
            result = list(yaml.safe_load_all(stream))

        if len(result) == 1:
            result = result[0]

        return result

    @classmethod
    def dump(cls, item: Any, stream: Optional[Union[IO, str, Path]] = None) -> str:
        """Dump an object to YAML following Grai's serialization rules

        Args:
            item: The object to dump
            stream: The stream to dump to. If None, the result is returned as a string.

        Returns:
            The YAML string

        Raises:

        """

        prepped_data = cls.prep_data(item)
        dumper: Callable = yaml.safe_dump_all if isinstance(item, Sequence) else yaml.safe_dump

        if stream is None:
            result = dumper(prepped_data)
        elif isinstance(stream, (str, Path)):
            with open(stream, "w") as file:
                result = dumper(prepped_data, file)
        else:
            result = dumper(prepped_data, stream)

        return result

    @staticmethod
    def prep_data(data: Any) -> Union[str, List[str]]:
        """Ensures that the data is in a format compliant with Grai's serialization rules.

        Args:
            data: The data to prepare

        Returns:
            If the data is a sequence, a list of strings is returned. Otherwise, a string is returned.

        Raises:

        """
        if isinstance(data, Sequence):
            return list(load_json(dump_json(item)) for item in data)
        else:
            return load_json(dump_json(data))



from typing import Callable, Dict, List, Literal, Optional, Type, Union
from uuid import UUID

from grai_schemas.v1.generics import GraiBaseModel, NamedID, UuidID
from grai_schemas.v1.metadata.edges import GenericEdgeMetadataV1, Metadata
from grai_schemas.v1.metadata.metadata import EdgeMetadataV1, GraiEdgeMetadataV1
from grai_schemas.v1.node import NodeIdTypes
from grai_schemas.v1.source import DataSourceMixin, DataSourcesMixin, SourceSpec
from pydantic import validator


class EdgeNamedID(NamedID):
    """Class definition of EdgeNamedID"""

    pass


class EdgeUuidID(UuidID):
    """Class definition of EdgeUuidID"""

    pass


EdgeIdTypes = Union[EdgeUuidID, EdgeNamedID]


class BaseSourcedEdgeSpec(GraiBaseModel):
    """Class definition of BaseSourcedEdgeSpec

    Attributes:
        display_name: An optional short form name for the edge
        source: The source node of the edge
        destination: The destination node of the edge
        is_active: Whether the edge is active or not
        workspace: The workspace the edge belongs to
        data_source: The data source which created this edge
        metadata: Metadata associated with the edge.

    """

    display_name: Optional[str]
    source: NodeIdTypes
    destination: NodeIdTypes
    is_active: Optional[bool] = True
    workspace: Optional[UUID]
    data_source: SourceSpec
    metadata: GraiEdgeMetadataV1 = GraiEdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"))

    @validator("metadata", always=True, pre=True)
    def validate_metadata(cls, v: Optional[Union[Dict, GraiEdgeMetadataV1]]) -> GraiEdgeMetadataV1:
        if isinstance(v, GraiEdgeMetadataV1):
            return v
        elif isinstance(v, dict):
            v.setdefault("grai", GenericEdgeMetadataV1(edge_type="Generic"))
            return GraiEdgeMetadataV1(**v)
        elif v is None:
            return GraiEdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"))
        raise ValueError(f"Invalid metadata: {v}. Expected either None, a dict, or a MetadataV1 instance.")

    def __str__(self) -> str:
        return f"Edge[Node({self.source}) -> Node({self.destination})]"


class NamedSourceSpec(EdgeNamedID, BaseSourcedEdgeSpec):
    """Class definition of NamedSourceSpec"""

    def to_edge(self) -> "NamedSpec":
        """

        Returns:
            A NamedSpec instance
        """
        values = self.dict(exclude={"data_source", "metadata"})
        values["data_sources"] = [self.data_source]
        values["metadata"] = {
            "grai": self.metadata.grai,
            "sources": {self.data_source.name: self.metadata},
        }
        return NamedSpec(**values)


class IDSourceSpec(EdgeUuidID, BaseSourcedEdgeSpec):
    """Class definition of IDSourceSpec"""

    def to_edge(self) -> "IDSpec":
        """

        Returns:
            An IDSpec instance
        """
        values = self.dict(exclude={"data_source", "metadata"})
        values["data_sources"] = [self.data_source]
        values["metadata"] = {
            "grai": self.metadata.grai,
            "sources": {self.data_source.name: self.metadata},
        }
        return IDSpec(**values)


SourcedEdgeSpec = Union[IDSourceSpec, NamedSourceSpec]


class SourcedEdgeV1(GraiBaseModel):
    """Class definition of SourcedEdgeV1

    Attributes:
        type: The type of the edge e.g. NodeV1, EdgeV1, etc...
        version: Object version e.g. v1
        spec: The edge specification

    """

    type: Literal["SourceEdge"]
    version: Literal["v1"]
    spec: SourcedEdgeSpec

    @classmethod
    def from_spec(cls, spec_dict: Dict) -> "SourcedEdgeV1":
        """

        Args:
            spec_dict (Dict):

        Returns:
            A SourcedEdgeV1 instance

        Raises:

        """
        return cls(version="v1", type="SourceEdge", spec=spec_dict)

    def __hash__(self):
        return hash(self.spec)

    def to_edge(self) -> "EdgeV1":
        """Converts a SourcedEdgeV1 instance to an EdgeV1 instance

        Returns:
            An EdgeV1 instance
        """
        return EdgeV1(version="v1", type="Edge", spec=self.spec.to_edge())


class BaseEdgeSpec(GraiBaseModel):
    """Class definition of BaseEdgeSpec

    Attributes:
        display_name: An optional short form name for the edge
        source: The source node of the edge
        destination: The destination node of the edge
        is_active: Whether the edge is active or not
        workspace: The workspace the edge belongs to
        data_sources: The data sources which have contributed to this edge
        metadata: Metadata associated with the edge.

    """

    display_name: Optional[str]
    source: NodeIdTypes
    destination: NodeIdTypes
    is_active: Optional[bool] = True
    workspace: Optional[UUID]
    data_sources: List[Union[UUID, SourceSpec]]
    metadata: EdgeMetadataV1 = EdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"), sources={})

    @validator("metadata", always=True, pre=True)
    def validate_metadata(cls, v: Optional[Union[Dict, EdgeMetadataV1]]) -> EdgeMetadataV1:
        if isinstance(v, EdgeMetadataV1):
            return v
        elif isinstance(v, dict):
            v.setdefault("grai", GenericEdgeMetadataV1(edge_type="Generic"))
            v.setdefault("sources", {})
            return EdgeMetadataV1(**v)
        elif v is None:
            return EdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"), sources={})
        raise ValueError(f"Invalid metadata: {v}. Expected either None, a dict, or a MetadataV1 instance.")

    def __str__(self):
        return f"Edge[Node({self.source}) -> Node({self.destination})]"


class NamedSpec(EdgeNamedID, BaseEdgeSpec):
    """Class definition of NamedSpec"""

    pass


class IDSpec(EdgeUuidID, BaseEdgeSpec):
    """Class definition of IDSpec"""

    pass


EdgeSpec = Union[IDSpec, NamedSpec]


class EdgeV1(GraiBaseModel):
    """Class definition of EdgeV1

    Attributes:
        type: The type of the edge e.g. NodeV1, EdgeV1, etc...
        version: Object version e.g. v1
        spec: The edge specification

    """

    type: Literal["Edge"]
    version: Literal["v1"]
    spec: EdgeSpec

    @classmethod
    def from_spec(cls, spec_dict: Dict) -> "EdgeV1":
        """

        Args:
            spec_dict (Dict):

        Returns:
            An EdgeV1 instance

        Raises:

        """
        return cls(version="v1", type="Edge", spec=spec_dict)

    def __hash__(self):
        return hash(self.spec)
